(* :Title: MyFindRoot *)
(* :Author: Silvio Levy *)
 
(* :Summary: Searches for a numerical solution of the equations with
more assurance than the built-in FindRoot function. *)  
 
(* :Context: MyFindRoot` *)
 
(* :Package Version: 1 *)
 
(* :Copyright: Copyright 1991 Silvio Levy. *)
 
(* :Mathematica Versions: 1.2 *)
 
(* :History:
  Separated from the package GeneralEquilibrium.m, which were supplied
  in the electronic supplement of The Mathematica Journal, Volume 1,
  Issue 3, Winter 1991.  GeneralEquilibrium.m was written by Asahi
  Noguchi in August 1990, modified by Silvio Levy in February/March
  1991.
*)

(* :Discussion: 
MyFindRoot is slower than FindRoot (though not by a large factor).
It always uses approximations to the derivative, rather than computing
derivatives symbolically.  Therefore the specification of variables
is of the form {x,x0} or {x,x0,xmin,xmax}, and never {x,{x0,x1}} as
with FindRoot.
*)
 
BeginPackage["MFRoot`"]
 
MyFindRoot::usage = "\n
  MyFindRoot is variant of FindRoot that doesn't give up when\n
  it lands out of bounds; rather, it uses various heuristics\n
  to try to recover and continue the search for a solution.\n
  It is used like FindRoot."
 
Begin["`private`"]
  
MyFindRoot::tomanyiter = "too many iterations."
MyFindRoot::wrongdim = "wrong number of equations or variables."
MyFindRoot::nonnum = "non-numerical result of equations."
 
SeedRandom[0]
Off[Plot::notnum]
Off[ParametricPlot::notpoint]
 
MyFindRoot[oeqs_,varsandopts__] :=
Block[
  {dim,vars,opts,ag,maxiter,eqs=oeqs,
    eps=.5 10^-6,derivs,image,i=0,j,sol,t,names,vals,code},
  vars=Cases[{varsandopts},_List];
  opts=Complement[{varsandopts},vars];
  ag=AccuracyGoal/.opts/.Options[FindRoot];
  If[SameQ[ag,Automatic], ag=eps, ag=10^-ag];
  maxiter=MaxIterations/.opts/.Options[FindRoot];
  If[!SameQ[Head[eqs],List],eqs=List[eqs]];
  eqs=eqs/.Equal[x_,y_]->x-y;
  dim=Length[vars];
  If[dim!=Length[eqs], Message[MyFindRoot::wrongdim]; Return[Null]];
  names=First /@ vars;
  vals=First/@(Rest/@ vars);
  bounds=Rest/@(Rest /@ vars);
  
  While[True,
    rules=Thread[names->vals];
    image=N[eqs/.rules];
    If[!(And@@(NumberQ/@image)),  Message[ MyFindRoot::nonnum];  Return[Null]];
    If[Max[Abs[image]]<ag, Return[rules]];
    If[i==maxiter,  Message[MyFindRoot::tomanyiter]; Return[rules,image]];
    derivs=Table[N[eqs/. MapAt[(#+eps)&,rules,{j,2}]]-image,{j,1,dim}];
    sol=vals - image.Inverse[derivs/eps];
    t=adjust[Min[timeout /@ Transpose[{sol,vals,bounds}]]];
    vals=Thread[jolt[t sol + (1-t)vals,bounds]];
    i++;
  ]
]
 
timeout[{newval_,curval_,{}}] := Infinity
timeout[{newval_,curval_,{lo_,hi_}}] :=
  If[newval<curval,(lo-curval)/(newval-curval),(curval-hi)/(curval-newval)]
 
adjust[t_] := If[t>4, 1, 2t/(4+t)]
 
jolt[val_?NumberQ,{}] := val
jolt[val_?NumberQ,{lo_,hi_}] := 
  If[val-lo< .001 || hi-val < .001, Random[Real,{.25(3lo+hi),.25(lo+3hi)}],val]
 
End[]
EndPackage[]
